@@ -8,9 +8,11 @@ package Mail::JMAPTalk;
 use HTTP::Tiny;
 use JSON::XS qw(decode_json encode_json);
 use Convert::Base64;
+use File::LibMagic;
 use Carp qw(confess);
+use Data::Dumper;
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 our $CLIENT = "Net-JMAPTalk";
 our $AGENT = "$CLIENT/$VERSION";
 
@@ -36,6 +38,18 @@ sub auth_header {
   return 'Basic ' . encode_base64("$Self->{user}:$Self->{password}", '');
 }
 
+sub uploaduri {
+  my $Self = shift;
+  my $scheme = $Self->{scheme} // 'http';
+  my $host = $Self->{host} // 'localhost';
+  my $port = $Self->{port} // ($scheme eq 'http' ? 80 : 443);
+  my $url = $Self->{uploadurl} // '/upload';
+
+  return $url if $url =~ m/^http/;
+
+  return "$scheme://$host:$port$url";
+}
+
 sub uri {
   my $Self = shift;
   my $scheme = $Self->{scheme} // 'http';
@@ -99,6 +113,54 @@ sub Request {
   my $jdata;
   $jdata = eval { decode_json($Response->{content}) } if $Response->{success};
 
+  if ($ENV{DEBUGJMAP}) {
+    warn "JMAP " . Dumper($Requests, $Response);
+  }
+
+  # check your own success on the Response object
+  if (wantarray) {
+    return ($Response, $jdata);
+  }
+
+  confess "JMAP request for $Self->{user} failed ($uri): $Response->{status} $Response->{reason}: $Response->{content}"
+    unless $Response->{success};
+
+  confess "INVALID JSON $Response->{content}" unless $jdata;
+
+  return $jdata;
+}
+
+sub _get_type {
+  my $data = shift;
+  # XXX - escape file names?
+  my $magic = File::LibMagic->new();
+  my $info = $magic->info_from_string($data);
+  return $info->{mime_type};
+}
+
+sub Upload {
+  my ($Self, $data, $type) = @_;
+
+  my %Headers;
+  $Headers{'Content-Type'} = $type || _get_type($data);
+
+  if ($Self->{user}) {
+    $Headers{'Authorization'} = $Self->auth_header();
+  }
+  if ($Self->{token}) {
+    $Headers{'Authorization'} = "JMAP $Self->{token}";
+  }
+
+  my $uri = $Self->uploaduri();
+
+  my $Response = $Self->ua->post($uri, {
+    headers => \%Headers,
+    content => $data,
+  });
+
+  my $jdata;
+  $jdata = eval { decode_json($Response->{content}) } if $Response->{success};
+
   # check your own success on the Response object
   if (wantarray) {
     return ($Response, $jdata);
